"""
Numba implementation of Gringorten projection.

Author: Matthew Petroff <ORCID:0000-0002-4436-4215>
Year: 2020

Based on the Gringorten projection implementation in d3-geo-projection, which
is licensed under a 3-Clause BSD License:
https://github.com/d3/d3-geo-projection/blob/6b86866e19ba9959bb0ac115689ffac5ccf6d3db/LICENSE
"""

import numpy as np
import numba


@numba.njit
def gringorten_hexadecant(lambda_, phi):
    EPS = 1e-12

    sinPhi = np.sin(phi)
    r = sinPhi * sinPhi
    r2 = r * r
    j = 1 + r2
    k = 1 + 3 * r2
    q = 1 - r2
    z = np.arcsin(1 / np.sqrt(j))
    v = q + r * j * z
    p2 = (1 - sinPhi) / v
    p = np.sqrt(p2)
    a2 = p2 * j
    a = np.sqrt(a2)
    h = p * q

    if lambda_ == 0:
        return 0, -(h + r * a)

    cosPhi = np.cos(phi)
    secPhi = 1 / cosPhi
    drdPhi = 2 * sinPhi * cosPhi
    dvdPhi = (-3 * r + z * k) * drdPhi
    dp2dPhi = (-v * cosPhi - (1 - sinPhi) * dvdPhi) / (v * v)
    dpdPhi = (0.5 * dp2dPhi) / p
    dhdPhi = q * dpdPhi - 2 * r * p * drdPhi
    dra2dPhi = r * j * dp2dPhi + p2 * k * drdPhi
    mu = -secPhi * drdPhi
    nu = -secPhi * dra2dPhi
    zeta = -2 * secPhi * dhdPhi
    lambda1 = 4 * lambda_ / np.pi

    if lambda_ > 0.222 * np.pi or phi < np.pi / 4 and lambda_ > 0.175 * np.pi:
        # Slower but accurate bisection method
        x = (h + r * np.sqrt(a2 * (1 + r2) - h * h)) / (1 + r2)
        if lambda_ > np.pi / 4:
            return x, x
        x1 = x
        x0 = 0.5 * x
        x = 0.5 * (x0 + x1)
        for _ in range(50):
            g = np.sqrt(a2 - x * x)
            f = (x * (zeta + mu * g) + nu * np.arcsin(x / a)) - lambda1
            if not f:
                break
            if f < 0:
                x0 = x
            else:
                x1 = x
            x = 0.5 * (x0 + x1)
            if np.abs(x1 - x0) <= EPS * EPS:
                break
    else:
        # Newton-Raphson
        x = EPS * EPS
        for _ in range(25):
            x2 = x * x
            g2 = np.sqrt(a2 - x2)
            zetaMug = zeta + mu * g2
            f2 = x * zetaMug + nu * np.arcsin(x / a) - lambda1
            df = zetaMug + (nu - mu * x2) / g2
            delta = f2 / df if g2 else 0
            x -= delta
            if np.abs(delta) <= EPS * EPS:
                break

    return x, -h - r * np.sqrt(a2 - x * x)


@numba.njit
def gringorten_quincuncial(lambda_, phi):
    lambda_ = lambda_ % (2 * np.pi)
    quadrant = lambda_ // (np.pi / 2)
    lambda_ = lambda_ % (np.pi / 2) - np.pi / 4
    x, y = gringorten_hexadecant(np.abs(lambda_), np.abs(phi))
    x *= np.sign(lambda_) / 2
    y = (np.sign(phi) * (y + 1) - 1) / 2
    cosrot = 1 - 2 * (((quadrant + 1) % 4) // 2)
    sinrot = 1 - 2 * (quadrant // 2)
    x_t = x * cosrot - y * sinrot
    y_t = x * sinrot + y * cosrot
    return x_t, y_t


def gringorten(lambda_, phi):
    return np.array(
        [gringorten_quincuncial(lambda_[i], phi[i]) for i in range(lambda_.shape[0])]
    ).T


@numba.njit
def gringorten_hexadecant_inverse(x, y):
    EPS = 1e-7

    x0 = 0
    x1 = 1
    r = 0.5
    i = 50

    while True:
        r2 = r * r
        sinPhi = np.sqrt(r)
        z = np.arcsin(1 / np.sqrt(1 + r2))
        v = (1 - r2) + r * (1 + r2) * z
        p2 = (1 - sinPhi) / v
        p = np.sqrt(p2)
        a2 = p2 * (1 + r2)
        h = p * (1 - r2)
        g2 = a2 - x * x
        g = np.sqrt(g2)
        y0 = y + h + r * g
        i -= 1
        if abs(x1 - x0) < EPS * EPS or i == 0 or y0 == 0:
            break
        if y0 > 0:
            x0 = r
        else:
            x1 = r
        r = 0.5 * (x0 + x1)

    if not i:
        return 0, 0

    phi = np.arcsin(sinPhi)
    cosPhi = np.cos(phi)
    secPhi = 1 / cosPhi
    drdPhi = 2 * sinPhi * cosPhi
    dvdPhi = (-3 * r + z * (1 + 3 * r2)) * drdPhi
    dp2dPhi = (-v * cosPhi - (1 - sinPhi) * dvdPhi) / (v * v)
    dpdPhi = 0.5 * dp2dPhi / p
    dhdPhi = (1 - r2) * dpdPhi - 2 * r * p * drdPhi
    zeta = -2 * secPhi * dhdPhi
    mu = -secPhi * drdPhi
    nu = -secPhi * (r * (1 + r2) * dp2dPhi + p2 * (1 + 3 * r2) * drdPhi)

    return np.pi / 4 * (x * (zeta + mu * g) + nu * np.arcsin(x / np.sqrt(a2))), phi


@numba.njit
def gringorten_quincuncial_inverse(x, y):
    quadrant = 0 + (y >= 0) + (x <= 0) + 2 * (y <= 0) * (x < 0)
    cosrot = 1 - 2 * (quadrant // 2)
    sinrot = 2 * (((quadrant - 1) % 4) // 2) - 1
    x_t = x * sinrot + y * cosrot
    y_t = x * cosrot - y * sinrot - 1
    y_t2 = np.abs(y_t) - 1

    lambda_, phi = gringorten_hexadecant_inverse(x_t, y_t2)

    lambda_ += np.pi / 4 + quadrant * np.pi / 2
    lambda_ %= 2 * np.pi
    return lambda_, -phi * np.sign(y_t)


def gringorten_inverse(x, y):
    return np.array(
        [gringorten_quincuncial_inverse(x[i], y[i]) for i in range(x.shape[0])]
    ).T
